//  Licensed under the Apache License, Version 2.0 (the "License"); you may not
//  use this file except in compliance with the License. You may obtain a copy of
//  the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
//  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
//  License for the specific language governing permissions and limitations under
//  the License.
//
//  ThemeFramework.m
//  iThemeFramework
//
//  Created by Chris Davis on 05/04/2012.
//  Copyright (c) 2012 GameWeaver LTD. All rights reserved.
//

#import "ThemeFramework.h"
#import "Constants.h"
#import "Networking.h"

@implementation ThemeFramework

static ThemeFramework* _sharedMySingleton = nil;

#pragma mark - Initialization methods

+ (ThemeFramework *)instance
{
	@synchronized([ThemeFramework class])
	{
		if (!_sharedMySingleton)
		{
			_sharedMySingleton = [[super alloc] init];
		}
		return _sharedMySingleton;
	}
    
	return nil;
}

- (id)init
{
	self = [super init];
	if (self) {
		self.ImageLookup = [[NSMutableDictionary alloc] init];
	}
	return self;
}

- (id)init:(NSString *)appKey themeKey:(NSString *)themeKey username:(NSString *)username password:(NSString *)password
{
	self = [super init];
	if (self) {
		self.AppKey = appKey;
		self.ThemeKey = themeKey;
		self.Username = username;
		self.Password = password;
		self.PlayerId = nil;
		self.AuthToken = nil;
		
		NSDictionary *playerDict = @{
		kUSERNAME: self.Username,
		kPASSWORD: self.Password
		};
		
		[self login:playerDict
			 caller:self start:^(id self) {
				 
			 } end:^(id self, NSDictionary *result, NSError *error) {
				 
			 }];
		
		dispatch_queue_t queue = dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0);
		dispatch_source_t timer = dispatch_source_create(DISPATCH_SOURCE_TYPE_TIMER, 0, 0, queue);
		dispatch_time_t now = dispatch_walltime(DISPATCH_TIME_NOW, 0);
		dispatch_source_set_timer(timer, now, 1ull*NSEC_PER_SEC, 5000ull); //Once a second
		dispatch_source_set_event_handler(timer, ^{
			[self updateLayoutEvent];
		});
		dispatch_resume(timer);
		
	}
	return self;
}

- (void)ensureLayout:(UIViewController *)controller
{
	[self ensureLayout:controller toInterfaceOrientation:UIInterfaceOrientationPortrait];
}

- (void)ensureLayout:(UIViewController *)controller toInterfaceOrientation:(UIInterfaceOrientation)toInterfaceOrientation
{
	self.Controller = controller;
	
	CGSize size = [self computeScreenSize];
	NSString *screenId = [self computeScreenName:controller size:size];
	
	//ensure screen is created within theme on server
	[self ensureScreen:controller screenId:screenId size:size];
	
	//TODO:
	//Check recursion on UIView's
	//Caching/CALayers/Images
	
	//TODO: Take screenshot & upload
	[self takeAndUploadScreenshot:controller];
	
	//swizzle methods
	[self swizzleMethods];
}

- (void)updateLayoutEvent
{
	if (self.AuthToken == nil)
		return;
	
	CGSize size = [self computeScreenSize];
	NSString *screenId = [self computeScreenName:self.Controller size:size];
	NSString *url = [NSString stringWithFormat:kAPI_READ_SCREEN(@"en-gb", self.PlayerId, self.AppKey, self.ThemeKey, screenId)];
	
	NSError *error = nil;
	Networking *net = [[Networking alloc] init];
	NSDictionary *result = [net httpSynchronousUrlWithMethod:url
													  method:kHTTP_GET
												   keyvalues:nil
														auth:self.AuthToken
													   error:&error];
	[net release];
	
	if (result == nil)
		return;
	
	dispatch_async(dispatch_get_main_queue(), ^{
		[self applyLayoutTo:self.Controller.view with:result];
	});

}

- (void)ensureScreen:(UIViewController *)controller screenId:(NSString *)screenId size:(CGSize)size
{
	
	if (self.AuthToken == nil)
		return;
	
	NSArray *children = [self walkControls:controller.view];
	
	NSDictionary *layout = @{
	@"test": @"ok",
	@"width": NSStringFromNSInteger((int)size.width),
	@"height": NSStringFromNSInteger((int)size.height),
	@"children": children
	};
	
	NSString *url = [NSString stringWithFormat:kAPI_READ_SCREEN(@"en-gb", self.PlayerId, self.AppKey, self.ThemeKey, screenId)];
	
	//Upload Layout.
	NSError *error = nil;
	Networking *net = [[Networking alloc] init];
	NSDictionary *result = [net httpSynchronousUrlWithMethod:url
													  method:kHTTP_PUT
												   keyvalues:layout
														auth:self.AuthToken
													   error:&error];
	[net release];
}

- (NSArray *)walkControls:(UIView *)view
{
	//List of class types we can iterate, feel free to add to.
	NSArray *whiteList = @[[UIView class], [UIButton class], [UILabel class]];
	
	//DLog(@"--%@", NSStringFromClass([view class]));
	NSMutableArray *items = [[NSMutableArray alloc] init];
	for (UIView *subView in view.subviews)
	{
		//Skip if not in the whitelist.
		if ([whiteList containsObject:[subView class]] == FALSE)
			continue;
		
		NSArray *children = [self walkControls:subView];
		//DLog(@"%@", NSStringFromClass([subView class]));
		
		NSDictionary *ctrl = @{
		@"tag":NSStringFromNSInteger(subView.tag),
		@"children":children,
		@"kind":NSStringFromClass([subView class]),
		@"frame": NSStringFromCGRect(subView.frame),
		@"bgcolor": [self colorToWeb:subView.backgroundColor],
		@"bgimage": @"",
		@"bgimage_url": @""
		};
		
		[items addObject:ctrl];
	}
	
	return items;
}

- (NSArray *)applyLayoutTo:(UIView *)view with:(NSDictionary *)dic
{
	//List of class types we can iterate, feel free to add to.
	NSArray *whiteList = @[[UIView class], [UIButton class], [UILabel class]];
	
	//DLog(@"--%@", NSStringFromClass([view class]));
	
	NSMutableArray *d = [dic objectForKey:@"children"];
	
	NSMutableArray *items = [[NSMutableArray alloc] init];
	int i = 0;
	for (UIView *subView in view.subviews)
	{
		//Skip if not in the whitelist.
		if ([whiteList containsObject:[subView class]] == FALSE)
			continue;
		
		NSArray *children = [self walkControls:subView];
		
		NSDictionary *item = [d objectAtIndex:i];
		NSInteger tag = [[item objectForKey:@"tag"] integerValue];
		CGRect frame = CGRectFromString([item objectForKey:@"frame"]);
		NSString *bgcolor = [item objectForKey:@"bgcolor"];
		UIColor *bguicolor = subView.backgroundColor;
		if (bgcolor)
			subView.backgroundColor = [self colorFromHexString:bgcolor];
		
		//DLog(@"%@", NSStringFromClass([subView class]));
		
		if (subView.tag == tag)
			subView.frame = frame;
		
		i++;
	}
	
	return items;
}

- (CGSize)computeScreenSize
{
	CGRect screenBounds = [[UIScreen mainScreen] bounds];
	CGFloat screenScale = [[UIScreen mainScreen] scale];
	CGSize size = CGSizeMake(screenBounds.size.width * screenScale, screenBounds.size.height * screenScale);
	if (UIInterfaceOrientationIsLandscape([UIApplication sharedApplication].statusBarOrientation)) {
        size = CGSizeMake(size.height, size.width);
    }
	return size;
}

- (NSString *)computeScreenName:(UIViewController *)controller size:(CGSize)size
{
	return [NSString stringWithFormat:@"%@_%d_%d", NSStringFromClass([controller class]), (int)size.width, (int)size.height];
}

- (void)takeAndUploadScreenshot:(UIViewController *)controller
{
	UIImage *image = [self takeScreenshot];
	
	if (image == nil)
		return;
	
	
	NSData *data = UIImagePNGRepresentation(image);
	
	//TODO: Get upload Url
	NSString *uploadUrl = @"";
	
	//Upload to retreived upload Url
	NSError *error = nil;
	Networking *net = [[Networking alloc] init];
	NSDictionary *result = [net postImage:uploadUrl
								imageData:data
									 auth:self.AuthToken
									error:&error];
	[net release];
}

//Sometimes we cannot take a screenshot if
//the view hasn't been initalized.
- (UIImage *)takeScreenshot
{
	UIWindow *keyWindow = [[UIApplication sharedApplication] keyWindow];
	if (keyWindow == nil)
		return nil;
	
	CGRect rect = [keyWindow bounds];
	UIGraphicsBeginImageContextWithOptions(rect.size, NO, 0);
	CGContextRef context = UIGraphicsGetCurrentContext();
	[keyWindow.layer renderInContext:context];
	UIImage *img = UIGraphicsGetImageFromCurrentImageContext();
	CGImageRef imageRef = CGImageCreateWithImageInRect([img CGImage], rect);
	UIGraphicsEndImageContext();
	return [UIImage imageWithCGImage:imageRef];
}

// More methods can be added here at will.
- (void)swizzleMethods
{
	[self swizzleClassMethodToInstanceMethod:[UIImage class] instance:[self class] orig:@selector(imageNamed:) newmethod:@selector(_imageNamed:)];
}

- (void)swizzleClassMethodToInstanceMethod:(Class)c instance:(id)instance orig:(SEL)orig newmethod:(SEL)newmethod
{
    Method origMethod = class_getClassMethod(c, orig);
    Method newMethod = class_getInstanceMethod(instance, newmethod);
    if(class_addMethod(c, orig, method_getImplementation(newMethod), method_getTypeEncoding(newMethod)))
        class_replaceMethod(c, newmethod, method_getImplementation(origMethod), method_getTypeEncoding(origMethod));
    else
		method_exchangeImplementations(origMethod, newMethod);
}

- (UIImage *)_imageNamed:(NSString *)name
{
	return [[_sharedMySingleton ImageLookup] objectForKey:name];
}

- (void)downloadAndCacheImage:(NSString *)urlString
{
	UIImage *img = [self.ImageLookup objectForKey:urlString];
	if (img != nil)
	{
		return;
	}
	
	dispatch_queue_t concurrentQueue = dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0);
	dispatch_async(concurrentQueue, ^{
		
		__block UIImage *image = nil;
		
		dispatch_sync(concurrentQueue, ^{
			
			NSURL *url = [NSURL URLWithString:urlString];
			NSURLRequest *urlRequest = [NSURLRequest requestWithURL:url cachePolicy:NSURLRequestReturnCacheDataElseLoad timeoutInterval:60];
			NSError *error = nil;
			NSData *imageData = [NSURLConnection sendSynchronousRequest:urlRequest
													  returningResponse:nil
																  error:&error];
			if (error == nil && imageData != nil)
			{
				image = [UIImage imageWithData:imageData];
			} else if (error != nil)
			{
				//error
			} else {
				//no data from url
			}
		});
		
		dispatch_sync(dispatch_get_main_queue(), ^{
			if (image != nil)
			{
				[self.ImageLookup setObject:image forKey:urlString];
				
			} else {
				
			}
		});
		
		
	});
}

//Do not pass # into this.
- (UIColor *)colorFromHexString:(NSString *)hexString
{
    unsigned rgbValue = 0;
    NSScanner *scanner = [NSScanner scannerWithString:hexString];
    [scanner scanHexInt:&rgbValue];
    return [UIColor colorWithRed:((rgbValue & 0xFF0000) >> 16)/255.0 green:((rgbValue & 0xFF00) >> 8)/255.0 blue:(rgbValue & 0xFF)/255.0 alpha:1.0];
}

- (void)login:(NSDictionary *)playerDict
	   caller:(id)caller
		start:(StartBlock)start
		  end:(EndBlock)end
{
	if (start)
	{
		start(caller);
	}
	
	dispatch_queue_t concurrentQueue = dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0);
	dispatch_async(concurrentQueue, ^{
		
		NSError *error = nil;
		
		Networking *post = [[Networking alloc] init];
		NSDictionary *dic = [post httpSynchronousUrlWithMethod:kAPI_LOGIN(@"en-gb")
														method:kHTTP_POST
													 keyvalues:playerDict
														  auth:nil
														 error:&error];
		[post release];
		
		if (error != nil)
		{
			//Do something
		}
		
		
		int code =  [[dic objectForKey:@"code"] intValue];
		if (code == 1)
		{
			//Set AuthToken/PlayerId if they are returned.
			self.AuthToken =[dic objectForKey:@"authToken"];
			self.PlayerId =[dic objectForKey:@"playerId"];
		}
		
		if (end)
		{
			end(caller, dic, error);
		}
		
	});
}

- (NSString*)colorToWeb:(UIColor*)color
{
    NSString *webColor = @"";
	
    // This method only works for RGB colors
    if (color &&
        CGColorGetNumberOfComponents(color.CGColor) == 4)
    {
        // Get the red, green and blue components
        const CGFloat *components = CGColorGetComponents(color.CGColor);
		
        // These components range from 0.0 till 1.0 and need to be converted to 0 till 255
        CGFloat red, green, blue;
        red = roundf(components[0] * 255.0);
        green = roundf(components[1] * 255.0);
        blue = roundf(components[2] * 255.0);
		
        // Convert with %02x (use 02 to always get two chars)
        webColor = [[NSString alloc]initWithFormat:@"%02x%02x%02x", (int)red, (int)green, (int)blue];
    }
	
    return webColor;
}

@end
